/*
Dataprobe Exchange Protocol (DxP) control example
Release 1.2  03/19/2010

1.2 : 03/19/2010
                Add inputs to query command
                Change default to Channel 1 for both Control and Query

1.1 : 02/19/2010 Initial Public Release

This program queries the inputs and queries or controls the outputs for any Dataprobe
product that supports the DxP protocol.  (iBoot-G2, iPIO-8, etc.)  It is intended to provide
an example of how to utilize the protocol to control the products programatically.

This code is compiled to dxp_1.2.exe and distributed from Dataprobe's website at
http://dataprobe.com/files/control/dxp_1.2.zip     Includes this file, compiled
code and protocol documentation.

This source code and compiled program are distributed as an example and Dataprobe
makes no warranty as to its suitability for any purpose. It is freely distributable.
Feel free to make modifications and post on Dataprobe's support forum.

Support is available at dataprobe.com/forum/ tech@dataprobe.com  or 201-934-5111

*/

//// INCLUDES
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <winsock2.h>

//// DEFINES
#define NUL			0x00
#define VER			"v1.01.004"			// version of this program
#define DXP_VER		"v1.2"				// version of the DxP protocol this program supports

//// GLOBAL VARIABLES
struct header {
	unsigned char command;
	char uName[21];
	char password[21];
	unsigned char desc;
	unsigned char param;
	unsigned short seq;
} tHeader;

struct pulsePayload {
	char channel;
	char state;
	unsigned short pulseWidth;
} tPulse;

struct controlPayload {
	char channel;
	char state;
} tControl;

//// FUNCTIONS
void usage(void)
{
	printf("\nDxP Control Utility %s\n\n", VER);
	printf("Usage:\n");
	printf("  -a<ipaddress> - required - xxx.xxx.xxx.xxx \n");
	printf("  -n<ipport>    - optional - default is 9100 \n");
	printf("  -c<0-8>       - optional - default is 1, 0 is all 8 \n");
	printf("  -s<status>    - required (unless query) - 0 or 1  \n");
	printf("  -t<time>      - optional - cycle time 01 - 99 sec.  \n");
	printf("  -q<i|o>       - query (i)nput or (o)utput channel -c \n");
	printf("  -u<user>      - required on supported devices \n");
	printf("  -p<password>  - required on supported devices  \n");
	printf("  -v            - version information \n");
	exit(8);
}

int main(int argc, char *argv[])
{
	// Program entry point
	// initialize parameters for arguments and set defaults
	char command[128];					// DxP command buffer	
	char ipaddress[20] = " ";			// ip address from command line -a
	char hello[12];						// hello string is hello-000 null terminated
	char statword[10];					// temporary status buffer
	char recvbuf[5];					// receive buffer
	char channels[] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08};
	char chan = channels[1];			// 1 is the default
	char cmd0 = 0x01;					// 0x01 for control 0x04 for query outputs  0x06 query inputs 0x07 for pulse
	char st;							// relay state from command line -s
	char eCommand = 0x03, byte1, byte2;
	char aok = 0, sok = 0, sel = 1;		// boolean vars

	int ipport = 9100;					// 9100 is the default DxP Port -n
	int channel = 1;					// channel from command line -c
	int i, iResult, bytesSent, bytesRecv;
	
	unsigned short time;				// pulse time from command line -t
	unsigned short newseq;
	
	WSADATA wsaData;					// for Winsock
	SOCKET m_socket;
	struct sockaddr_in clientService;

	// setup initial hello message
	strcpy(hello, "hello-000");
	
	// zero out the command buffer
	memset(command, 0, sizeof(command));
	
	// zero out the header structure
	memset((char *)&tHeader, 0, sizeof(tHeader));

	// scan command line arguments
	while((argc > 1) && (argv[1][0] == '-'))
	{
		switch(argv[1][1])
		{
			case 'h':		// help
				usage();
				break;

			case 'a':		// ipaddress
				aok = 1;
				strcpy(ipaddress, &argv[1][2]);		// look at atoi(argv[i]+2)
				break;

			case 's':		// status
				sok = 1;
				switch(argv[1][2])
				{
					case '0':
						st = 0x00;
						break;
						
					case '1':
						st = 0x01;
						break;
					
					default:
						printf("State Error, must be 1 or 0");
						exit(5);
				}  // end switch status
				break;
				
			case 'q':		// do query
				sok = 1;
				switch(argv[1][2])
				{
					case 'i':
						cmd0 = 0x06;
						strcpy(statword, "Input ");
						break;
						
					case 'o':
						cmd0 = 0x04;
						strcpy(statword, "Output ");
						break;
						
					default:
						printf("Query Error, must be i or o");
						exit(5);
				}// end switch
				break;

			case 't':		// pulse command
				cmd0 = 0x07;
				time = atoi(&argv[1][2]);
				break;

			case 'n':		// port number
				ipport = atoi(&argv[1][2]);
				break;

			case 'c':		// channel select
				channel = atoi(&argv[1][2]);
				chan = channels[channel];
				if(channel == 0) {
					sel = 0;
				}
				break;

			case 'u':		// username
				if(strlen(&argv[1][2]) <= 20) {
					strcpy(tHeader.uName, &argv[1][2]);
				}
				else {
					printf("User name must be 20 characters or less\n");
					exit(5);
				}
				break;

			case 'p':		// password
				if(strlen(&argv[1][2]) <= 20) {
					strcpy(tHeader.password, &argv[1][2]);
				}
				else {
					printf("Password must be 20 characters or less\n");
					exit(5);
				}
				break;
				
			case 'v':		// version
				printf("DxP Control Utility %s\n", VER);
				printf("DxP Protocol %s\n", DXP_VER);
				exit(1);
				break;

			default:		// any other argument not defined
				printf("Wrong Argument: %s \n", argv[1]);
				usage();

		} // end switch

		++argv;
		--argc;
	}  //end while

	// check for -a and -s (or -q) parameters
	if(!aok || !sok){
		printf(" -a and ( -s or -q ) parameters required \n");
		usage();
	}

	// parameters done, lets do some networking....

	// Initialize Winsock.
	iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
	if(iResult != NO_ERROR) {
		printf("Client: Error at WSAStartup().\n");
		return 0;
	}

	// Create a socket.
	m_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if(m_socket == INVALID_SOCKET) {
		printf("Client: socket() - Error at socket(): %ld\n", WSAGetLastError());
		closesocket(m_socket);
		WSACleanup();
		return 0;
	}

	// Connect to the DxP server device.
	clientService.sin_family = AF_INET;
	clientService.sin_addr.s_addr = inet_addr(ipaddress);
	clientService.sin_port = htons(ipport);
	if(connect(m_socket, (SOCKADDR*)&clientService, sizeof(clientService)) == SOCKET_ERROR) {
		printf("Failed to connect to %s \n",ipaddress);
		closesocket(m_socket);
		WSACleanup();
		return 0;
	}
	else {
		printf("Connection made to %s \n", ipaddress);
	}

	// Send and receive data.

	// Sends Hello to Server
	bytesSent = send(m_socket, hello, 10, 0);
	if(bytesSent == SOCKET_ERROR) {
		printf("Client: send() error %ld.\n", WSAGetLastError());
		closesocket(m_socket);
		return 0;
	}

	// Receives Sequence from Server
	bytesRecv = recv(m_socket, recvbuf, 2, 0);
	if((bytesRecv == 0) || (bytesRecv == WSAECONNRESET)) {
		printf("Sequence not Returned - Fail \n");
		closesocket(m_socket);
		return 0;
	}

	// convert sequence # to Sequence +1
	byte1 = recvbuf[0];
	byte2 = recvbuf[1];

	newseq = (unsigned short)byte2;
	newseq <<= 8;			// shift to left 8 bits
	newseq |= byte1;		// added lower 8 bits of sequence #
	newseq++;				// increment to get new sequence #

	// build command line and send
	tHeader.command = eCommand;
	tHeader.desc = cmd0;
	tHeader.param = NUL;
	tHeader.seq = newseq;
		
	// copy tHeader to the command buffer
	memcpy(command, (char *)&tHeader, sizeof(tHeader));

	// build the payloads
	// pulse payload
	tPulse.channel = chan;
	tPulse.state = st;
	tPulse.pulseWidth = time;
		
	// control payload
	tControl.channel = chan;
	tControl.state = st;
	
	// combine header and payload and send.  Based on Command Type
	switch(cmd0) {
		case 0x01:     // control     send header and control payload
			memcpy(&command[sizeof(tHeader)], (char *)&tControl, sizeof(tControl));
			bytesSent = send(m_socket, command, 49, 0);
			break;

		case 0x04:      // query Outputs      just send header - no payload
		case 0x06:      // query Inputs      just send header - no payload
			bytesSent = send(m_socket, (char*)&tHeader, sizeof(tHeader), 0);

			// receive status and print out
			bytesRecv = recv(m_socket, recvbuf, 8, 0);
			if((bytesRecv == 0) || (bytesRecv == WSAECONNRESET)) {
				printf("Connection Closed by %s\n", ipaddress);
				closesocket(m_socket);
				return 0;
			}

			if(bytesRecv < 0)
				return 0;
			else {
				for(i = 0; i < 8; i++) {
					byte1 = recvbuf[i];
					if(!sel ||(sel && i+1 == channel)) {
						if(byte1 == 0x00) {
							printf("%s %d is Open / Off \n", statword, i+1);
						}
						else {
							printf("%s %d is Closed / On \n", statword, i+1);
						}
					}
					else {
						// do nothing
					}
				} //end for

			}//end if
			
			closesocket(m_socket);
			WSACleanup();
			return 0;

		case 0x07:      // pulse   send header and pulse data
			// combine header and payload
			memcpy(&command[sizeof(tHeader)], (char *)&tPulse, sizeof(tPulse));
			bytesSent = send(m_socket, command, 51, 0);
			break;

		default:
			printf("Program Error - Fail \n");
			return 0;
	}// end switch

	// receive response 0x00 is OK
	bytesRecv = recv(m_socket, recvbuf, 2, 0);
	if((bytesRecv == 0) || (bytesRecv == WSAECONNRESET)) {
		printf("Connection Closed by %s\n",ipaddress);
		closesocket(m_socket);
		WSACleanup();
		return 0;
	}

	if(bytesRecv < 0) {
		closesocket(m_socket);
		WSACleanup();
		return 0;
	}
	else {
		byte1 = recvbuf[0];
		if(byte1 == 0x00) {
			printf("ACK Received, OK \n");
		}
		else {
			printf("NAK received,  Fail \n");
		}
	}

	// We are done.  Close socket
	closesocket(m_socket);
	Sleep(200);                  // delay to allow socket to close
	WSACleanup();
	return 0;
}

